import Revolt from 'revolt.js';
import Discord from 'discord.js';
import Express from 'express';
import { config } from 'dotenv';
import { exec } from 'child_process';

config();

const { DISCORD_TOKEN, REVOLT_TOKEN, DISCORD_CHANNEL, REVOLT_CHANNEL, PORT, RESTART_COMMAND_BOT, RESTART_COMMAND_BRIDGE, RESTART_WEBHOOK } = process.env;

if (!DISCORD_TOKEN) throw '$DISCORD_TOKEN not set';
if (!REVOLT_TOKEN) throw '$REVOLT_TOKEN not set';
if (!DISCORD_CHANNEL) throw '$DISCORD_CHANNEL not set';
if (!REVOLT_CHANNEL) throw '$REVOLT_CHANNEL not set';
if (!PORT) throw '$PORT not set';

const TIMEOUT = 5;

const clients = {
    discord: new Discord.Client({ intents: [ 'MessageContent', 'GuildMessages' ] }),
    revolt: new Revolt.Client({ autoReconnect: true, pongTimeout: 30, onPongTimeout: 'RECONNECT' }),
}

let currentValue: Awaited<ReturnType<typeof periodicCheck>>|null = null;
let restarting = { bot: false, bridge: false };

const app = Express();
app.listen(PORT, () => console.log('Listening on :' + PORT));

clients.discord.login(DISCORD_TOKEN);
clients.revolt.useExistingSession({ token: REVOLT_TOKEN });

clients.discord.on('ready', () => console.log('Discord ready'));
clients.revolt.on('ready', () => console.log('Revolt ready'));

app.get('/healthcheck', async (req: Express.Request, res: Express.Response) => {
    console.log('Health check request received');

    if (!currentValue) {
        console.log('No health check result stored yet, running checks');
        await periodicCheck();
    }

    res.status(currentValue!.status).send(currentValue);
});

const periodicCheck = async () => {
    console.log('Running periodic check');
    const now = new Date().toISOString();
    const [ botStatus, bridgeStatusDiscord, bridgeStatusRevolt ] = await Promise.all([
        retry(checkBot, 3),
        retry(checkBridgeDiscord, 3),
        retry(checkBridgeRevolt, 3),
    ]);

    if (botStatus) restarting.bot = false;
    if (bridgeStatusDiscord && bridgeStatusRevolt) restarting.bridge = false;

    const statusCode = (botStatus && bridgeStatusDiscord && bridgeStatusRevolt) ? 200 : 503;
    const response = {
        status: statusCode,
        checkedOn: now,
        ready: {
            discord: !!clients.discord.readyAt,
            revolt: !!clients.revolt.websocket.ready,
        },
        watchdog: {
            bot: botStatus,
            bridgeDiscord: bridgeStatusDiscord,
            bridgeRevolt: bridgeStatusRevolt,
        },
    };

    currentValue = response;
    runRestarts(response).catch(e => console.log(e)); // Explicitly do not await this
    return response;
}

const whClient = RESTART_WEBHOOK ? new Discord.WebhookClient({ url: RESTART_WEBHOOK }) : null;
const runRestarts = async (status: Awaited<ReturnType<typeof periodicCheck>>) => {

    if (!status.watchdog.bot && !restarting.bot && RESTART_COMMAND_BOT) {
        console.log('Restarting bot');
        try {
            restarting.bot = true;
            const result = exec(RESTART_COMMAND_BOT!);
            result.stdout?.on('data', data => console.log(data));
            result.stderr?.on('data', data => console.log(data));
            result.on('exit', async (code) => {
                if (code != 0) {
                    console.log('Bot: Restart command seems to have failed!');
                    restarting.bot = false;
                    await whClient?.send(`Bot restart command failed with status ${code}`);
                }
                else await whClient?.send(`Bot restart command finished successfully!`);
            });

            await whClient?.send('Restarting bot');
        } catch(e) {
            console.log(e);
        }
    }

    if ((!status.watchdog.bridgeDiscord || !status.watchdog.bridgeRevolt) && !restarting.bridge && RESTART_COMMAND_BRIDGE) {
        console.log('Restarting bridge');
        try {
            restarting.bridge = true;
            const result = exec(RESTART_COMMAND_BRIDGE!);
            result.stdout?.on('data', data => console.log(data));
            result.stderr?.on('data', data => console.log(data));
            result.on('exit', async (code) => {
                if (code != 0) {
                    console.log('Bridge: Restart command seems to have failed!');
                    restarting.bridge = false;
                    await whClient?.send(`Bridge restart command failed with status ${code}`);
                }
                else await whClient?.send(`Bridge restart command finished successfully!`);
            });

            await whClient?.send('Restarting bridge');
        } catch(e) {
            console.log(e);
        }
    }
}

setInterval(async () => {
    if (clients.discord.readyAt && clients.revolt.user) {
        await periodicCheck();
    }
    else console.log('Cannot run periodic check as clients aren\'t ready yet');
}, 60000);

const checkBot = (): Promise<boolean> => new Promise(async (resolve, _reject) => {
    const id = Math.round(Math.random() * 1000000);
    console.log('Running health check for bot: ' + id);

    try {
        const channel = clients.revolt.channels.get(REVOLT_CHANNEL);
        
        const timeout = setTimeout(() => {
            clients.revolt.removeListener('message', cb);
            resolve(false);
            console.log('Health check timed out: ' + id);
        }, 1000 * TIMEOUT);
        
        const cb = (message: Revolt.Message) => {
            if (message.channel_id == REVOLT_CHANNEL &&
                message.author_id == '01FHGJ3NPP7XANQQH8C2BE44ZY' &&
                message.content == 'Health check success: ' + id
            ) {
                clients.revolt.removeListener('message', cb);
                resolve(true);
                clearTimeout(timeout);
                console.log('Health check succeeded: ' + id);
            }
        }
        
        clients.revolt.on('message', cb);
        await channel?.sendMessage('/healthcheck ' + id);
    } catch(e) {
        return false;
    }
});

// Send message on Discord, check if it arrives on Revolt
const checkBridgeRevolt = (): Promise<boolean> => new Promise(async (resolve, _reject) => {
    const id = Math.round(Math.random() * 1000000);
    console.log('Running health check for bridge on Revolt: ' + id);

    try {
        const channel = await clients.discord.channels.fetch(DISCORD_CHANNEL) as Discord.TextChannel;

        const timeout = setTimeout(() => {
            clients.revolt.removeListener('message', cb);
            resolve(false);
            console.log('Health check timed out: ' + id);
        }, 1000 * TIMEOUT);
        
        const cb = (message: Revolt.Message) => {
            if (message.channel_id == REVOLT_CHANNEL &&
                message.author_id == '01FHGJ3NPP7XANQQH8C2BE44ZY' &&
                message.content == '(Discord) Bridge health check: ' + id
            ) {
                clients.revolt.removeListener('message', cb);
                resolve(true);
                clearTimeout(timeout);
                console.log('Health check succeeded: ' + id);
            }
        }
        
        clients.revolt.on('message', cb);
        await channel?.send('(Discord) Bridge health check: ' + id);
    } catch(e) {
        return false;
    }
});

// Send message on Revolt, check if it arrives on Discord
const checkBridgeDiscord = (): Promise<boolean> => new Promise(async (resolve, _reject) => {
    const id = Math.round(Math.random() * 1000000);
    console.log('Running health check for bridge on Discord: ' + id);

    try {
        const channel = clients.revolt.channels.get(REVOLT_CHANNEL);

        const timeout = setTimeout(() => {
            clients.discord.removeListener('message', cb);
            resolve(false);
            console.log('Health check timed out: ' + id);
        }, 1000 * TIMEOUT);
        
        const cb = (message: Discord.Message) => {
            if (message.channelId == DISCORD_CHANNEL && message.webhookId && message.content == '(Revolt) Bridge health check: ' + id) {
                clients.revolt.removeListener('message', cb);
                resolve(true);
                clearTimeout(timeout);
                console.log('Health check succeeded: ' + id);
            }
        }
        
        clients.discord.on('messageCreate', cb);
        await channel?.sendMessage('(Revolt) Bridge health check: ' + id);
    } catch(e) {
        return false;
    }
});

const retry = (action: () => Promise<boolean>, retryCount: number) => new Promise(async (resolve, reject) => {
    try {
        for (let i = 0; i < retryCount; i++) {
            if (await action()) {
                resolve(true);
                return;
            }
        }
        resolve(false);
    } catch(e) {
        reject(e);
    }
});
